unit uPlaybackQueue;

interface

uses
  Windows, ActiveX, AIMPSDKAddons, Classes, Generics.Collections;

type

  { TAIMPStorageEntryInfo }

  TAIMPStorageEntryInfo = class(TObject)
  public
    EntryIndex: Integer;
    FileName: UnicodeString;
    Rating: Integer;
    RepeatCount: Integer;
    Statistics: TAIMPFileStatisticsInfo;
  end;

  { TAIMPStorageEntryInfoList }

  TAIMPStorageEntryInfoList = class(TObjectList<TAIMPStorageEntryInfo>)
  public
    function Find(const AEntryIndex: Integer; AStartFromIndex: Integer = 0): Integer; overload;
    function Find(const AFileName: UnicodeString; out AInfo: TAIMPStorageEntryInfo): Boolean; overload;
  end;

  { TAIMPStorageInfo }

  TAIMPStorageInfo = class(TObject)
  strict private
    FEntries: TAIMPStorageEntryInfoList;
    FEntriesAdded: TAIMPStorageEntryInfoList;
    FEntriesRemoved: TAIMPStorageEntryInfoList;
    FManager: IAIMPAddonsPlaylistManager;
    FStorage: HPLS;
    function StorageGetEntry(AEntryIndex: Integer): HPLSENTRY;
    function StorageGetEntryFileName(AEntryIndex: Integer): UnicodeString;
    procedure StorageGetEntryStatistics(AEntryIndex: Integer; var AStatistics: TAIMPFileStatisticsInfo);
  public
    constructor Create(AManager: IAIMPAddonsPlaylistManager); virtual;
    destructor Destroy; override;
    procedure Update;
    //
    property Entries: TAIMPStorageEntryInfoList read FEntries;
    property EntriesAdded: TAIMPStorageEntryInfoList read FEntriesAdded;
    property EntriesRemoved: TAIMPStorageEntryInfoList read FEntriesRemoved;
    property Storage: HPLS read FStorage write FStorage;
  end;

  { TAIMPShufflePlaybackQueue }

  TAIMPShufflePlaybackQueue = class(TObject)
  strict private
    FCursor: Integer;
    FItems: TAIMPStorageEntryInfoList;
    FManager: IAIMPAddonsPlaylistManager;
    FStorageInfo: TAIMPStorageInfo;
    function GetStorage: HPLS;
    procedure SetCursor(AValue: Integer);
    procedure SetStorage(AValue: HPLS);
  protected
    FSuggestedEntryIndex: Integer;
    FSuggestedItemIndex: Integer;
    procedure CalculateItemsRating;
    procedure CalculateItemsRepeatCount;
    procedure ItemsAdd(AItems: TAIMPStorageEntryInfoList);
    procedure ItemsRemove(AItems: TAIMPStorageEntryInfoList);

    property Cursor: Integer read FCursor write SetCursor;
    property Items: TAIMPStorageEntryInfoList read FItems;
    property StorageInfo: TAIMPStorageInfo read FStorageInfo;
  public
    constructor Create(AManager: IAIMPAddonsPlaylistManager); virtual;
    destructor Destroy; override;
    function GetFile(var AEntryIndex: Integer; ANext: Boolean): Boolean;
    procedure UpdateContent;
    procedure UpdateCursor; overload;
    procedure UpdateCursor(APlayingEntryIndex: Integer); overload;
    //
    property Storage: HPLS read GetStorage write SetStorage;
  end;

implementation

uses
{$IFDEF DEBUG}
  uDebugMonitor,
{$ENDIF}
  SysUtils, AIMPAddonCustomPlugin, Math;

{ TAIMPStorageEntryInfoList }

function TAIMPStorageEntryInfoList.Find(const AEntryIndex: Integer; AStartFromIndex: Integer = 0): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := AStartFromIndex to Count - 1 do
    if AEntryIndex = List[I].EntryIndex then
    begin
      Result := I;
      Break;
    end;
end;

function TAIMPStorageEntryInfoList.Find(const AFileName: UnicodeString; out AInfo: TAIMPStorageEntryInfo): Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to Count - 1 do
  begin
    Result := SameText(AFileName, List[I].FileName);
    if Result then
    begin
      AInfo := List[I];
      Break;
    end;
  end;
end;

{ TAIMPStorageInfo }

constructor TAIMPStorageInfo.Create(AManager: IAIMPAddonsPlaylistManager);
begin
  inherited Create;
  FManager := AManager;
  FEntries := TAIMPStorageEntryInfoList.Create;
  FEntriesAdded := TAIMPStorageEntryInfoList.Create(False);
  FEntriesRemoved := TAIMPStorageEntryInfoList.Create(False);
end;

destructor TAIMPStorageInfo.Destroy;
begin
  FreeAndNil(FEntriesRemoved);
  FreeAndNil(FEntriesAdded);
  FreeAndNil(FEntries);
  inherited Destroy;
end;

procedure TAIMPStorageInfo.Update;

  function GetFiles(AHandle: HPLS): TStringList;
  var
    AEntriesCount: Integer;
    AEntryIndex: Integer;
  begin
    Result := TStringList.Create;
    AEntriesCount := FManager.StorageGetEntryCount(AHandle);
    if AEntriesCount > 0 then
    begin
      Result.Capacity := AEntriesCount;
      for AEntryIndex := 0 to AEntriesCount - 1 do
        Result.Add(StorageGetEntryFileName(AEntryIndex));
    end;
  end;

  procedure FetchAddedFiles(AFiles: TStringList);
  var
    AItem: TAIMPStorageEntryInfo;
    I: Integer;
  begin
    EntriesAdded.Clear;
    for I := 0 to AFiles.Count - 1 do
    begin
      if not Entries.Find(AFiles[I], AItem) then
      begin
        AItem := TAIMPStorageEntryInfo.Create;
        AItem.FileName := AFiles[I];
        StorageGetEntryStatistics(I, AItem.Statistics);
        EntriesAdded.Add(AItem);
        Entries.Add(AItem);
      end;
      AItem.EntryIndex := I;
    end;
  end;

  procedure FetchRemovedFiles(AFiles: TStringList);
  var
    I: Integer;
  begin
    EntriesRemoved.Clear;
    for I := Entries.Count - 1 downto 0 do
      if AFiles.IndexOf(Entries[I].FileName) < 0 then
      begin
        EntriesRemoved.Add(Entries[I]);
        Entries.Delete(I);
      end;
  end;

var
  AFiles: TStringList;
begin
  AFiles := GetFiles(Storage);
  try
    FetchRemovedFiles(AFiles);
    FetchAddedFiles(AFiles);
  finally
    AFiles.Free;
  end;
end;

function TAIMPStorageInfo.StorageGetEntry(AEntryIndex: Integer): HPLSENTRY;
begin
  Result := FManager.StorageGetEntry(Storage, AEntryIndex);
end;

function TAIMPStorageInfo.StorageGetEntryFileName(AEntryIndex: Integer): UnicodeString;
begin
  Result := EntryPropertyGetStringValue(FManager, StorageGetEntry(AEntryIndex), AIMP_PLAYLIST_ENTRY_PROPERTY_FILENAME);
end;

procedure TAIMPStorageInfo.StorageGetEntryStatistics(AEntryIndex: Integer; var AStatistics: TAIMPFileStatisticsInfo);
begin
  if Failed(FManager.EntryPropertyGetValue(StorageGetEntry(AEntryIndex),
    AIMP_PLAYLIST_ENTRY_PROPERTY_STATISTICS, @AStatistics, SizeOf(AStatistics)))
  then
    ZeroMemory(@AStatistics, SizeOf(AStatistics));
end;

{ TAIMPShufflePlaybackQueue }

constructor TAIMPShufflePlaybackQueue.Create(AManager: IAIMPAddonsPlaylistManager);
begin
  inherited Create;
  FManager := AManager;
  FSuggestedItemIndex := -1;
  FSuggestedEntryIndex := -1;
  FItems := TAIMPStorageEntryInfoList.Create(False);
  FStorageInfo := TAIMPStorageInfo.Create(AManager);
  Randomize;
end;

destructor TAIMPShufflePlaybackQueue.Destroy;
begin
  FreeAndNil(FStorageInfo);
  FreeAndNil(FItems);
  inherited Destroy;
end;

function TAIMPShufflePlaybackQueue.GetFile(var AEntryIndex: Integer; ANext: Boolean): Boolean;
begin
  FSuggestedItemIndex := -1;
  if ANext then
  begin
    if Cursor + 1 < Items.Count then
      FSuggestedItemIndex := Cursor + 1;
  end
  else
    if Cursor > 0 then
      FSuggestedItemIndex := Cursor - 1;

  Result := FSuggestedItemIndex >= 0;
  if Result then
  begin
    FSuggestedEntryIndex := Items[FSuggestedItemIndex].EntryIndex;
    AEntryIndex := FSuggestedEntryIndex;
  end
  else
    FSuggestedEntryIndex := -1;
end;

procedure TAIMPShufflePlaybackQueue.UpdateContent;
begin
  StorageInfo.Update;
  CalculateItemsRating;
  CalculateItemsRepeatCount;
  ItemsRemove(StorageInfo.EntriesRemoved);
  ItemsAdd(StorageInfo.EntriesAdded);
{$IFDEF DEBUG}
  TShuffleDebugMonitorForm.Update(Items, Cursor);
{$ENDIF}
end;

procedure TAIMPShufflePlaybackQueue.UpdateCursor;
var
  AIndex: Integer;
begin
  if Succeeded(FManager.StoragePropertyGetValue(Storage, AIMP_PLAYLIST_STORAGE_PROPERTY_PLAYINGINDEX, @AIndex, SizeOf(AIndex))) then
  begin
    if AIndex >= 0 then
      UpdateCursor(AIndex);
  end;
end;

procedure TAIMPShufflePlaybackQueue.UpdateCursor(APlayingEntryIndex: Integer);
var
  AIndex: Integer;
begin
  if FSuggestedEntryIndex = APlayingEntryIndex then
    Cursor := FSuggestedItemIndex
  else
  begin
    AIndex := Items.Find(APlayingEntryIndex, Cursor + 1);
    if AIndex < 0 then
      AIndex := Items.Find(APlayingEntryIndex, 0);
    if (AIndex >= 0) and (AIndex <> Cursor) then
    begin
      if AIndex > Cursor then
        Cursor := Cursor + 1;
      Items.Move(AIndex, Cursor);
      FSuggestedEntryIndex := APlayingEntryIndex;
      FSuggestedItemIndex := Cursor;
    end;
  end;
{$IFDEF DEBUG}
  TShuffleDebugMonitorForm.Update(Items, Cursor);
{$ENDIF}
end;

procedure TAIMPShufflePlaybackQueue.CalculateItemsRating;

  function GetAddingTimeOfFirstFile: TDateTime;
  var
    I: Integer;
  begin
    Result := Now;
    for I := 0 to StorageInfo.Entries.Count - 1 do
      Result := Min(Result, StorageInfo.Entries[I].Statistics.AddingTime);
  end;

  function GetTotalFilePlayedCount: Integer;
  var
    I: Integer;
  begin
    Result := 0;
    for I := 0 to StorageInfo.Entries.Count - 1 do
      Inc(Result, StorageInfo.Entries[I].Statistics.PlayCount);
  end;

  function CalculateRating(const AFileStatistics: TAIMPFileStatisticsInfo;
    const AAddingTimeOfFirstFile, ACurrentTime: TDateTime; ATotalFilePlayedCount: Integer): Integer;
  const
    PrecisionFactor = 10;
  var
    ADenominator: Double;
  begin
    if ATotalFilePlayedCount > 0 then
    begin
      ADenominator := Power(ATotalFilePlayedCount,
        1 - (AFileStatistics.AddingTime - AAddingTimeOfFirstFile) / (2 * (ACurrentTime - AAddingTimeOfFirstFile)));
    end
    else
      ADenominator := 1;

    Result := Round(PrecisionFactor * AFileStatistics.PlayCount / ADenominator);
  end;

var
  AAddingTimeOfFirstFile: TDateTime;
  ACurrentTime: TDateTime;
  AEntry: TAIMPStorageEntryInfo;
  ATotalFilePlayedCount: Integer;
  I: Integer;
begin
  ACurrentTime := Now;
  AAddingTimeOfFirstFile := GetAddingTimeOfFirstFile;
  ATotalFilePlayedCount := GetTotalFilePlayedCount;

  for I := 0 to StorageInfo.Entries.Count - 1 do
  begin
    AEntry := StorageInfo.Entries[I];
    AEntry.Rating := CalculateRating(AEntry.Statistics, AAddingTimeOfFirstFile, ACurrentTime, ATotalFilePlayedCount);
  end;
end;

procedure TAIMPShufflePlaybackQueue.CalculateItemsRepeatCount;
var
  AEntry: TAIMPStorageEntryInfo;
  I, AMinRatingValue: Integer;
begin
  if StorageInfo.Entries.Count > 0 then
  begin
    AMinRatingValue := MaxInt;
    for I := 0 to StorageInfo.Entries.Count - 1 do
    begin
      AMinRatingValue := Min(AMinRatingValue, StorageInfo.Entries[I].Rating);
      if AMinRatingValue = 0 then
        Break;
    end;
    for I := 0 to StorageInfo.Entries.Count - 1 do
    begin
      AEntry := StorageInfo.Entries[I];
      AEntry.RepeatCount := 1 + AEntry.Rating - AMinRatingValue;
    end;
  end;
end;

procedure TAIMPShufflePlaybackQueue.ItemsAdd(AItems: TAIMPStorageEntryInfoList);
var
  AEntryInfo: TAIMPStorageEntryInfo;
  ATempList: TAIMPStorageEntryInfoList;
  I, J: Integer;
begin
  if AItems.Count > 0 then
  begin
    ATempList := TAIMPStorageEntryInfoList.Create(False);
    try
      ATempList.Capacity := 3 * StorageInfo.EntriesAdded.Count;

      // Expand new items depended on it rating
      for I := 0 to StorageInfo.EntriesAdded.Count - 1 do
      begin
        AEntryInfo := StorageInfo.EntriesAdded[I];
        for J := 0 to AEntryInfo.RepeatCount - 1 do
          ATempList.Add(AEntryInfo);
      end;

      // Insert new items to the Queue
      while ATempList.Count > 0 do
      begin
        J := Random(ATempList.Count);
        Items.Insert(RandomRange(Cursor + 1, Items.Count + 1), ATempList[J]);
        ATempList.Delete(J);
      end;
    finally
      ATempList.Free;
    end;
  end;
end;

procedure TAIMPShufflePlaybackQueue.ItemsRemove(AItems: TAIMPStorageEntryInfoList);
var
  I, AIndex: Integer;
begin
  for I := 0 to AItems.Count - 1 do
  begin
    repeat
      AIndex := Items.Remove(AItems[I]);
      if (AIndex >= 0) and (AIndex <= Cursor) then
        Cursor := Cursor - 1;
    until AIndex = -1;
  end;
end;

function TAIMPShufflePlaybackQueue.GetStorage: HPLS;
begin
  Result := StorageInfo.Storage;
end;

procedure TAIMPShufflePlaybackQueue.SetCursor(AValue: Integer);
begin
  FCursor := Max(-1, AValue);
end;

procedure TAIMPShufflePlaybackQueue.SetStorage(AValue: HPLS);
begin
  if Storage <> AValue then
  begin
    FSuggestedItemIndex := -1;
    FSuggestedEntryIndex := -1;
    StorageInfo.Storage := AValue;
    Cursor := -1;
    UpdateContent;
    UpdateCursor;
  end;
end;

end.
